/*************************************************************************************
 Data Visualization - 3D Tree Map Demo                
 Tested in Processing 3                                    
 
 Author: Rodger (Jieliang) Luo                        
 Supervisor: George Legrady                           
 
 Purpose: Showing how to present multiple tree maps in 3D space by 
 using curve lines as connections.      
 
 Usage: 1. A mouse left-drag will rotate the camera around the subject.
 2. A right drag will zoom in and out. 
 3. A middle-drag (command-left-drag on mac) will pan. 
 4. A double-click restores the camera to its original position. 
 5. The shift key constrains rotation and panning to one axis or the other.
 
 Required Libraries:
 1. Treemap Library by Ben Fry
 2. Peasy Cam -> Accessible from Processing Import Library 
 3. Control P5 -> Accessible from Processing Import Library
 *************************************************************************************/

import treemap.*; //Tree Map library
import peasy.*; //3D Navigation library
import controlP5.*; //GUI library

//Constants Section
final int NUM_LAYERS = 12;   //Number of Layers
final int LAYER_DIST = 200;  //Distance between each layer
final int TRANS = 50;

int selectedDewey = -100;

boolean presentationMode = true;

//Camera(3D Nav) varaible
PeasyCam cam;

ControlP5 cp5;

//Data-Related varaibles
Table table;
Table deweyTable;
int numRows, numColumns;
float[][] dataMatrix;
String[] deweys;
ArrayList<String> deweyNames;

//TreeMap-Related varaibles
DeweyMap[] deweyData;
Treemap[] map;

ArrayList<ArrayList<PVector>> pointsOne = new ArrayList<ArrayList<PVector>>(NUM_LAYERS);
ArrayList<ArrayList<PVector>> pointsTwo = new ArrayList<ArrayList<PVector>>(NUM_LAYERS);
ArrayList<ArrayList<PVector>> pointsThree = new ArrayList<ArrayList<PVector>>(NUM_LAYERS);
ArrayList<ArrayList<PVector>> pointsFour = new ArrayList<ArrayList<PVector>>(NUM_LAYERS);

void setup() {
  size(1280, 720, OPENGL);
  cam = new PeasyCam(this, 2000);

  cp5 = new ControlP5(this);

  Textfield Dewey = cp5.addTextfield("dewey")
                       .setPosition(160, 10)
                       .setSize(50, 15)
                       .setFocus(true)
                       .setColor(color(0))
                       .setColorBackground(color(255))
                       .setColorActive(color(255))
                       . setColorForeground(color(255))
                       .setCaptionLabel("Please Enter A Dewey Number from 0 to 990")
                       ;
  Dewey.getCaptionLabel().getStyle().marginTop  = -16; 
  Dewey.getCaptionLabel().getStyle().marginLeft  = 55;
    
  cp5.addCheckBox("checkBox")
     .setPosition(20, 10)
     .setSize(15, 15) 
     .addItem("Presentation Mode", 1)
     .activateAll()
     ;
     
  cp5.setAutoDraw(false);

  //Put data into Processing table from csv
  table = loadTable("2006.csv", "header");
  numRows = table.getRowCount(); 
  //numRows = 1;
  numColumns = table.getColumnCount();
  dataMatrix = new float[numColumns-1][numRows];
  deweys = new String[numRows];
  println("Rows: " + numRows + " Columns: " + numColumns);
  
  deweyTable = loadTable("deweyClasses.csv", "header");
  deweyNames = new ArrayList<String>();
  for(int i=0; i<deweyTable.getRowCount(); i++){
    if(deweyTable.getInt(i,0) % 10 == 0){
      deweyNames.add(deweyTable.getString(i,1)); 
    }
  }
  
  //Put data from Table to 2D array
  for (int i=0; i<numColumns-1; i++) {
    for (int j=0; j<numRows; j++) {
      dataMatrix[i][j] = table.getFloat(j, i+1);
    }
  }

  //Save Dewey categories into 1D array
  for (int i=0; i<numRows; i++) {
    deweys[i] = table.getString(i, 0);
  }

  //Init tree map varaibles
  deweyData = new DeweyMap[NUM_LAYERS];
  map = new Treemap[NUM_LAYERS];
  for (int i=0; i<NUM_LAYERS; i++) { 
    deweyData[i]= new DeweyMap();
    pointsOne.add(new ArrayList<PVector>());
    pointsTwo.add(new ArrayList<PVector>());
    pointsThree.add(new ArrayList<PVector>());
    pointsFour.add(new ArrayList<PVector>());
  }

  //Add dewey categories and counts into tree map data structure
  for (int i=0; i<dataMatrix.length; i++) {
    for (int j=0; j<dataMatrix[i].length; j++) {
      deweyData[i].addDewey(deweys[j], (int)dataMatrix[i][j]);
    }
    deweyData[i].finishAdd();
    map[i] = new Treemap(deweyData[i], 0, 0, width, height);
  }

  //Draw basic tree maps 
  for (int i=0; i<dataMatrix.length; i++) {
    //Find and save the central point of each rect
    for (int j=0; j<deweyData[i].getItems().length; j++) {
      double x = deweyData[i].getItems()[j].getBounds().x;
      double y = deweyData[i].getItems()[j].getBounds().y;
      double w = deweyData[i].getItems()[j].getBounds().w; 
      double h = deweyData[i].getItems()[j].getBounds().h; 
      //PVector centralPoint = new PVector((int)(x+w/2), (int)(y+h/2), LAYER_DIST*i);
      //centralPoints.get(i).add(centralPoint);
      PVector pointOne = new PVector((int)x, (int)y, LAYER_DIST*(i-dataMatrix.length/2));
      pointsOne.get(i).add(pointOne);

      PVector pointTwo = new PVector((int)(x+w), (int)y, LAYER_DIST*(i-dataMatrix.length/2));
      pointsTwo.get(i).add(pointTwo);

      PVector pointThree = new PVector((int)(x+w), (int)(y+h), LAYER_DIST*(i-dataMatrix.length/2));
      pointsThree.get(i).add(pointThree);

      PVector pointFour = new PVector((int)x, (int)(y+h), LAYER_DIST*(i-dataMatrix.length/2));
      pointsFour.get(i).add(pointFour);
    }
  }
}

void draw() {
  background(220);
  colorMode(HSB);

  if(presentationMode) {
    rotateY(frameCount*0.001);
    if(frameCount % 120 == 0) {
      selectedDewey += 10;
      if(selectedDewey == 1000) selectedDewey = -100;
    }
  }

  //Draw Top
  pushMatrix();
  translate(-width/2, -height/2);
  for (int j=0; j<numRows; j++) {
   for (int i=0; i<pointsOne.size()-1; i++) {
     float x1_1 = pointsOne.get(i).get(j).x;
     float y1_1 = pointsOne.get(i).get(j).y + 0.1;
     float z1_1 = pointsOne.get(i).get(j).z;

     float x2_1 = pointsTwo.get(i).get(j).x;
     float y2_1 = pointsTwo.get(i).get(j).y + 0.1;
     float z2_1 = pointsTwo.get(i).get(j).z;

     float x1_2 = pointsOne.get(i+1).get(j).x;
     float y1_2 = pointsOne.get(i+1).get(j).y + 0.1;
     float z1_2 = pointsOne.get(i+1).get(j).z;

     float x2_2 = pointsTwo.get(i+1).get(j).x;
     float y2_2 = pointsTwo.get(i+1).get(j).y + 0.1;
     float z2_2 = pointsTwo.get(i+1).get(j).z;

     float lineColor = Float.parseFloat(deweyData[i].deweys.keySet().toArray()[j].toString());

     beginShape();  
     noStroke();
     if (lineColor == selectedDewey) {
       fill(map(lineColor, 0, 990, 0, 255), 255/2, 255, TRANS*3);
     } else {
       if (selectedDewey >= 0) noFill();
       else fill(map(lineColor, 0, 990, 0, 255), 255/2, 255/2, TRANS);
     } 
     vertex(x1_1, y1_1, z1_1);
     vertex(x2_1, y2_1, z2_1);
     vertex(x2_2, y2_2, z2_2);
     vertex(x1_2, y1_2, z1_2);
     endShape();
   }
  }
  popMatrix();

  //Bottom
  pushMatrix();
  translate(-width/2, -height/2);
  for (int j=0; j<numRows; j++) {
   for (int i=0; i<pointsThree.size()-1; i++) {
     float x1_1 = pointsThree.get(i).get(j).x;
     float y1_1 = pointsThree.get(i).get(j).y - 0.1;
     float z1_1 = pointsThree.get(i).get(j).z;

     float x2_1 = pointsFour.get(i).get(j).x;
     float y2_1 = pointsFour.get(i).get(j).y - 0.1;
     float z2_1 = pointsFour.get(i).get(j).z;

     float x1_2 = pointsThree.get(i+1).get(j).x;
     float y1_2 = pointsThree.get(i+1).get(j).y - 0.1;
     float z1_2 = pointsThree.get(i+1).get(j).z;

     float x2_2 = pointsFour.get(i+1).get(j).x;
     float y2_2 = pointsFour.get(i+1).get(j).y - 0.1;
     float z2_2 = pointsFour.get(i+1).get(j).z;

     float lineColor = Float.parseFloat(deweyData[i].deweys.keySet().toArray()[j].toString());     

     beginShape();
     noStroke();
     if (lineColor == selectedDewey) {
       fill(map(lineColor, 0, 990, 0, 255), 255/2, 255, TRANS*3);
     } else {
       if (selectedDewey >= 0) noFill();
       else fill(map(lineColor, 0, 990, 0, 255), 255/2, 255/2, TRANS);
     } 
     vertex(x1_1, y1_1, z1_1);
     vertex(x2_1, y2_1, z2_1);
     vertex(x2_2, y2_2, z2_2);
     vertex(x1_2, y1_2, z1_2);
     endShape();
   }
  }
  popMatrix();

  //Left
  pushMatrix();
  translate(-width/2, -height/2);
  for (int j=0; j<numRows; j++) {
   for (int i=0; i<pointsOne.size()-1; i++) {
     float x1_1 = pointsOne.get(i).get(j).x + 0.1;
     float y1_1 = pointsOne.get(i).get(j).y;
     float z1_1 = pointsOne.get(i).get(j).z;

     float x2_1 = pointsFour.get(i).get(j).x + 0.1;
     float y2_1 = pointsFour.get(i).get(j).y;
     float z2_1 = pointsFour.get(i).get(j).z;

     float x1_2 = pointsOne.get(i+1).get(j).x + 0.1;
     float y1_2 = pointsOne.get(i+1).get(j).y;
     float z1_2 = pointsOne.get(i+1).get(j).z;

     float x2_2 = pointsFour.get(i+1).get(j).x + 0.1;
     float y2_2 = pointsFour.get(i+1).get(j).y;
     float z2_2 = pointsFour.get(i+1).get(j).z;

     float lineColor = Float.parseFloat(deweyData[i].deweys.keySet().toArray()[j].toString());     

     beginShape();
     noStroke();
     if (lineColor == selectedDewey) {
       fill(map(lineColor, 0, 990, 0, 255), 255/2, 255, TRANS*3);
     } else {
       if (selectedDewey >= 0) noFill();
       else fill(map(lineColor, 0, 990, 0, 255), 255/2, 255/2, TRANS);
     } 
     vertex(x1_1, y1_1, z1_1);
     vertex(x2_1, y2_1, z2_1);
     vertex(x2_2, y2_2, z2_2);
     vertex(x1_2, y1_2, z1_2);
     endShape();
   }
  }
  popMatrix();

  //Right
  pushMatrix();
  translate(-width/2, -height/2);
  for (int j=0; j<numRows; j++) {
   for (int i=0; i<pointsTwo.size()-1; i++) {
     float x1_1 = pointsTwo.get(i).get(j).x - 0.1;
     float y1_1 = pointsTwo.get(i).get(j).y;
     float z1_1 = pointsTwo.get(i).get(j).z;

     float x2_1 = pointsThree.get(i).get(j).x - 0.1;
     float y2_1 = pointsThree.get(i).get(j).y;
     float z2_1 = pointsThree.get(i).get(j).z;

     float x1_2 = pointsTwo.get(i+1).get(j).x - 0.1;
     float y1_2 = pointsTwo.get(i+1).get(j).y;
     float z1_2 = pointsTwo.get(i+1).get(j).z;

     float x2_2 = pointsThree.get(i+1).get(j).x - 0.1;
     float y2_2 = pointsThree.get(i+1).get(j).y;
     float z2_2 = pointsThree.get(i+1).get(j).z;

     float lineColor = Float.parseFloat(deweyData[i].deweys.keySet().toArray()[j].toString());     
     lineColor = map(lineColor, 0, 990, 0, 255);

     beginShape();
     noStroke();
     if (lineColor == selectedDewey) {
       fill(map(lineColor, 0, 990, 0, 255), 255/2, 255, TRANS*3);
     } else {
       if (selectedDewey >= 0) noFill();
       else fill(map(lineColor, 0, 990, 0, 255), 255/2, 255/2, TRANS);
     } 
     vertex(x1_1, y1_1, z1_1);
     vertex(x2_1, y2_1, z2_1);
     vertex(x2_2, y2_2, z2_2);
     vertex(x1_2, y1_2, z1_2);
     endShape();
   }
  }
  popMatrix();

  //Draw basic tree maps 
  for (int i=0; i<dataMatrix.length; i++) {
    pushMatrix();
    translate(-width/2, -height/2, LAYER_DIST*(i - dataMatrix.length/2));
    map[i].draw();
    popMatrix();
  }

  //Draw Title
  cam.beginHUD();  //Stopping peasy cam
  fill(150);
  rect(0, 0, width, 35);
  fill(255);
  textAlign(RIGHT);
  if(selectedDewey >= 0 && selectedDewey <= 990){
    String deweyName = deweyNames.get(selectedDewey/10);
    deweyName = deweyName.toUpperCase();
    text(selectedDewey + " " + deweyName, width-10, 20);
  }
  cp5.draw();
  cam.endHUD();
}

public void dewey(String theText) {
  // automatically receives results from controller input
  println("a textfield event for controller 'input' : "+theText);
  selectedDewey = int(theText);
}

void checkBox(float[] a) {
  if(a[0] == 1) presentationMode = true;
  else presentationMode = false;
}